#include <algorithm>
#include <fstream>
#include <iostream>
#include <stack>

#include <sdsl/int_vector.hpp>
#include <sdsl/util.hpp>

#include <rlcsa/bits/bitbuffer.h>
#include <rlcsa/misc/utils.h>

#include "doclist.h"

//--------------------------------------------------------------------------

Doclist::Doclist(const RLCSA& _rlcsa) :
  rlcsa(_rlcsa), rmq(0), status(_rlcsa.isOk() ? st_unfinished : st_error)
{
}

Doclist::~Doclist()
{
  delete this->rmq; rmq = 0;
}

Doclist::result_type*
Doclist::listDocuments(const std::string& pattern, found_type* found) const
{
  if(!(this->isOk())) { return 0; }
  pair_type range = this->rlcsa.count(pattern);
  if(CSA::isEmpty(range)) { return 0; }
  return this->listUnsafe(range, found);
}

Doclist::result_type*
Doclist::listDocuments(pair_type range, found_type* found) const
{
  if(!(this->isOk()) || CSA::isEmpty(range) || range.second >= this->rlcsa.getSize()) { return 0; }
  return this->listUnsafe(range, found);
}

Doclist::result_type*
Doclist::listUnsafe(pair_type range, found_type* found) const
{
  bool delete_found = (found == 0);
  if(delete_found) { found = new found_type(this->rlcsa.getNumberOfSequences(), 1); }

  std::stack<pair_type> intervals; intervals.push(this->getRMQRange(range));
  result_type* results = new result_type;
  while(!(intervals.empty()))
  {
    pair_type current = intervals.top(); intervals.pop();
    usint min_pos = (*(this->rmq))(current.first, current.second);
    usint doc = this->docAt(min_pos, range);
    if(!(found->isSet(doc)))
    {
      this->addDocs(doc, min_pos, results, found, range);
      if(min_pos < current.second) { intervals.push(pair_type(min_pos + 1, current.second)); }
      if(min_pos > current.first)  { intervals.push(pair_type(current.first, min_pos - 1)); }
    }
  }

  CSA::sequentialSort(results->begin(), results->end());
  if(delete_found) { delete found; found = 0; }
  else
  {
    for(result_type::iterator iter = results->begin(); iter != results->end(); ++iter)
    {
      found->unsetBit(*iter);
    }
  }
  return results;
}

//--------------------------------------------------------------------------

const std::string DoclistSada::EXTENSION = ".sada";

DoclistSada::DoclistSada(const RLCSA& _rlcsa) :
  Doclist(_rlcsa)
{
  if(this->status != st_unfinished || !(this->rlcsa.supportsLocate())) { return; }

  // Build the C array.
  usint* prev = new usint[this->rlcsa.getNumberOfSequences()];
  for(usint i = 0; i < this->rlcsa.getNumberOfSequences(); i++) { prev[i] = 0; }
  sdsl::int_vector<64> c_array(this->rlcsa.getSize());
  for(usint i = 0; i < this->rlcsa.getSize(); i += BLOCK_SIZE)
  {
    pair_type range(i, std::min(i + BLOCK_SIZE, this->rlcsa.getSize()) - 1);
    usint* docarray = this->rlcsa.locate(range);
    this->rlcsa.getSequenceForPosition(docarray, CSA::length(range));
    for(usint j = range.first; j <= range.second; j++)
    {
      c_array[j] = prev[docarray[j - i]];
      prev[docarray[j - i]] = j + 1;
    }
    delete[] docarray; docarray = 0;
  }
  delete[] prev; prev = 0;

  // Build the RMQ.
  this->rmq = new RMQType(&c_array);

  this->status = st_ok;
}

DoclistSada::DoclistSada(const RLCSA& _rlcsa, const std::string& base_name) :
  Doclist(_rlcsa)
{
  if(this->status != st_unfinished || !(this->rlcsa.supportsLocate())) { return; }

  std::string doclist_name = base_name + EXTENSION;
  std::ifstream input(doclist_name.c_str(), std::ios_base::binary);
  if(input)
  {
    this->loadFrom(input);
    input.close();
  }
}

DoclistSada::DoclistSada(const RLCSA& _rlcsa, std::ifstream& input) :
  Doclist(_rlcsa)
{
  if(this->status != st_unfinished || !(this->rlcsa.supportsLocate())) { return; }
  this->loadFrom(input);
}

DoclistSada::~DoclistSada()
{
}

void
DoclistSada::writeTo(const std::string& base_name) const
{
  if(!(this->isOk())) { return; }
  std::string doclist_name = base_name + EXTENSION;
  std::ofstream output(doclist_name.c_str(), std::ios_base::binary);
  if(output)
  {
    this->writeTo(output);
    output.close();
  }
}

void
DoclistSada::writeTo(std::ofstream& output) const
{
  if(!(this->isOk())) { return; }
  this->rmq->serialize(output);
}

void
DoclistSada::loadFrom(std::ifstream& input)
{
  this->rmq = new RMQType;
  this->rmq->load(input);
  this->status = st_ok;
}

usint
DoclistSada::reportSize() const
{
  usint rmq_bytes = (this->rmq == 0 ? 0 : sdsl::util::get_size_in_bytes(*(this->rmq)));
  return sizeof(*this) + rmq_bytes;
}

pair_type
DoclistSada::getRMQRange(pair_type sa_range) const
{
  return sa_range;
}

usint
DoclistSada::docAt(usint rmq_index, pair_type sa_range) const
{
  return this->rlcsa.getSequenceForPosition(this->rlcsa.locate(rmq_index));
}

void
DoclistSada::addDocs(usint first_doc, usint rmq_index, result_type* results, found_type* found, pair_type sa_range) const
{
  results->push_back(first_doc);
  found->setBit(first_doc);
}

//--------------------------------------------------------------------------

const std::string DoclistILCP::EXTENSION = ".ilcp";

DoclistILCP::DoclistILCP(const RLCSA& _rlcsa) :
  Doclist(_rlcsa), run_heads(0)
{
  if(this->status != st_unfinished || !(this->rlcsa.supportsLocate())) { return; }

  // Build ILCP
  uint* iLCP = new uint[this->rlcsa.getSize()];
  for(usint i = 0; i < this->rlcsa.getNumberOfSequences(); i++)
  {
    CSA::SuffixArray* sa = this->rlcsa.getSuffixArrayForSequence(i);
    uint* plcp = sa->getLCPArray(true);
    for(usint j = 0; j < sa->getSize(); j++)
    {
      iLCP[sa->getRanks()[j] - this->rlcsa.getNumberOfSequences()] = plcp[j];
    }
    delete[] plcp; plcp = 0;
    delete sa; sa = 0;
  }

  // Count run heads
  usint n_run_heads = 1;
  for(usint i = 1; i < this->rlcsa.getSize(); i++)
  {
    if(iLCP[i] != iLCP[i - 1]) { n_run_heads++; }
  }

  // Build RMQ for the run heads
  sdsl::int_vector<32> heads(n_run_heads); heads[0] = iLCP[0];
  CSA::DeltaVector::Encoder encoder(DV_BLOCK_SIZE); encoder.setBit(0);
  for(usint i = 1, j = 1; i < this->rlcsa.getSize(); i++)
  {
    if(iLCP[i] != iLCP[i - 1])
    {
      heads[j] = iLCP[i]; j++;
      encoder.setBit(i);
    }
  }
  this->rmq = new RMQType(&heads);
  this->run_heads = new CSA::DeltaVector(encoder, this->rlcsa.getSize());

  this->status = st_ok;
}

DoclistILCP::DoclistILCP(const RLCSA& _rlcsa, const std::string& base_name) :
  Doclist(_rlcsa), run_heads(0)
{
  if(this->status != st_unfinished || !(this->rlcsa.supportsLocate())) { return; }

  std::string doclist_name = base_name + EXTENSION;
  std::ifstream input(doclist_name.c_str(), std::ios_base::binary);
  if(input)
  {
    this->loadFrom(input);
    input.close();
  }
}

DoclistILCP::DoclistILCP(const RLCSA& _rlcsa, std::ifstream& input) :
  Doclist(_rlcsa), run_heads(0)
{
  if(this->status != st_unfinished || !(this->rlcsa.supportsLocate())) { return; }
  this->loadFrom(input);
}

DoclistILCP::~DoclistILCP()
{
  delete this->run_heads; this->run_heads = 0;
}

void
DoclistILCP::writeTo(const std::string& base_name) const
{
  if(!(this->isOk())) { return; }
  std::string doclist_name = base_name + EXTENSION;
  std::ofstream output(doclist_name.c_str(), std::ios_base::binary);
  if(output)
  {
    this->writeTo(output);
    output.close();
  }
}

void
DoclistILCP::writeTo(std::ofstream& output) const
{
  if(!(this->isOk())) { return; }
  this->rmq->serialize(output);
  this->run_heads->writeTo(output);
}

void
DoclistILCP::loadFrom(std::ifstream& input)
{
  this->rmq = new RMQType;
  this->rmq->load(input);
  this->run_heads = new CSA::DeltaVector(input);
  this->status = st_ok;
}

usint
DoclistILCP::reportSize() const
{
  usint rmq_bytes = (this->rmq == 0 ? 0 : sdsl::util::get_size_in_bytes(*(this->rmq)));
  usint run_head_bytes = (this->run_heads == 0 ? 0 : this->run_heads->reportSize());
  return sizeof(*this) + rmq_bytes + run_head_bytes;
}

pair_type
DoclistILCP::getRMQRange(pair_type sa_range) const
{
  CSA::DeltaVector::Iterator iter(*(this->run_heads));
  return pair_type(iter.rank(sa_range.first) - 1, iter.rank(sa_range.second) - 1);
}

usint
DoclistILCP::docAt(usint rmq_index, pair_type sa_range) const
{
  CSA::DeltaVector::Iterator iter(*(this->run_heads));
  usint sa_index = std::max(sa_range.first, iter.select(rmq_index));
  return this->rlcsa.getSequenceForPosition(this->rlcsa.locate(sa_index));
}

void
DoclistILCP::addDocs(usint first_doc, usint rmq_index, result_type* results, found_type* found, pair_type sa_range) const
{
  results->push_back(first_doc); found->setBit(first_doc);

  CSA::DeltaVector::Iterator iter(*(this->run_heads));
  pair_type range(std::max(sa_range.first, iter.select(rmq_index)) + 1, 0);
  range.second = std::min(sa_range.second, iter.selectNext() - 1);

  if(CSA::isEmpty(range)) { return; }

  usint* res = this->rlcsa.locate(range);
  if(res != 0)
  {
    this->rlcsa.getSequenceForPosition(res, CSA::length(range));
    for(usint i = range.first; i <= range.second; i++)
    {
      results->push_back(res[i - range.first]);
      found->setBit(res[i - range.first]);
    }
    delete[] res; res = 0;
  }
}

//--------------------------------------------------------------------------

const std::string DoclistMuthu::EXTENSION = ".muthu";

DoclistMuthu::DoclistMuthu(const RLCSA& _rlcsa) :
  Doclist(_rlcsa), docarray(0)
{
  if(this->status != st_unfinished || !(this->rlcsa.supportsLocate())) { return; }

  usint item_bits = std::max((usint)1, CSA::length(this->rlcsa.getNumberOfSequences() - 1));
  CSA::WriteBuffer buf(this->rlcsa.getSize(), item_bits);
  buf.goToItem(0);

  // Prepare the stuctures.
  usint* prev = new usint[this->rlcsa.getNumberOfSequences()];
  for(usint i = 0; i < this->rlcsa.getNumberOfSequences(); i++) { prev[i] = 0; }
  sdsl::int_vector<64> c_array(this->rlcsa.getSize());
  for(usint i = 0; i < this->rlcsa.getSize(); i += BLOCK_SIZE)
  {
    pair_type range(i, std::min(i + BLOCK_SIZE, this->rlcsa.getSize()) - 1);
    usint* docs = this->rlcsa.locate(range);
    this->rlcsa.getSequenceForPosition(docs, CSA::length(range));
    for(usint j = range.first; j <= range.second; j++)
    {
      buf.writeItem(docs[j - i]);
      c_array[j] = prev[docs[j - i]];
      prev[docs[j - i]] = j + 1;
    }
    delete[] docs; docs = 0;
  }
  delete[] prev; prev = 0;

  // Build the structures.
  this->rmq = new RMQType(&c_array);
  this->docarray = buf.getReadBuffer();

  this->status = st_ok;
}

DoclistMuthu::DoclistMuthu(const RLCSA& _rlcsa, const std::string& base_name) :
  Doclist(_rlcsa), docarray(0)
{
  if(this->status != st_unfinished || !(this->rlcsa.supportsLocate())) { return; }

  std::string doclist_name = base_name + EXTENSION;
  std::ifstream input(doclist_name.c_str(), std::ios_base::binary);
  if(input)
  {
    this->loadFrom(input);
    input.close();
  }
}

DoclistMuthu::DoclistMuthu(const RLCSA& _rlcsa, std::ifstream& input) :
  Doclist(_rlcsa), docarray(0)
{
  if(this->status != st_unfinished || !(this->rlcsa.supportsLocate())) { return; }
  this->loadFrom(input);
}

DoclistMuthu::~DoclistMuthu()
{
  delete this->docarray; this->docarray = 0;
}

void
DoclistMuthu::writeTo(const std::string& base_name) const
{
  if(!(this->isOk())) { return; }
  std::string doclist_name = base_name + EXTENSION;
  std::ofstream output(doclist_name.c_str(), std::ios_base::binary);
  if(output)
  {
    this->writeTo(output);
    output.close();
  }
}

void
DoclistMuthu::writeTo(std::ofstream& output) const
{
  if(!(this->isOk())) { return; }
  this->rmq->serialize(output);
  this->docarray->writeBuffer(output);
}

void
DoclistMuthu::loadFrom(std::ifstream& input)
{
  this->rmq = new RMQType;
  this->rmq->load(input);

  usint item_bits = std::max((usint)1, CSA::length(this->rlcsa.getNumberOfSequences() - 1));
  this->docarray = new CSA::ReadBuffer(input, this->rlcsa.getSize(), item_bits);

  this->status = st_ok;
}

usint
DoclistMuthu::reportSize() const
{
  usint rmq_bytes = (this->rmq == 0 ? 0 : sdsl::util::get_size_in_bytes(*(this->rmq)));
  usint docarray_bytes = (this->docarray == 0 ? 0 : this->docarray->reportSize());
  return sizeof(*this) + rmq_bytes + docarray_bytes;
}

pair_type
DoclistMuthu::getRMQRange(pair_type sa_range) const
{
  return sa_range;
}

usint
DoclistMuthu::docAt(usint rmq_index, pair_type sa_range) const
{
  return this->docarray->readItem(rmq_index);
}

void
DoclistMuthu::addDocs(usint first_doc, usint rmq_index, result_type* results, found_type* found, pair_type sa_range) const
{
  results->push_back(first_doc);
  found->setBit(first_doc);
}

//--------------------------------------------------------------------------
