#include <algorithm>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <vector>

#include "doclist.h"

#include <rlcsa/misc/utils.h>
#include <rlcsa/docarray.h>


void printUsage();
void printResults(const std::string& algorithm, double seconds, usint docc, usint patterns);
void printSize(const std::string& algorithm, usint text_size, usint index_size, usint doclist_size);
void writeRanges(std::vector<pair_type>& ranges, const std::string& file_name);

int
main(int argc, char** argv)
{
  std::cout << "Document listing test" << std::endl;
  if(argc < 3)
  {
    printUsage();
    return 1;
  }

  // Parameters
  char* base_arg = 0;
  char* pattern_arg = 0;
  bool brute = false, ilcp = false, muthu = false, pdl = false, sada = false, selected = false;
  for(int arg = 1; arg < argc; arg++)
  {
    if(argv[arg][0] == '-')
    {
      switch(argv[arg][1])
      {
        case 'b':
          brute = true; selected = true; break;
        case 'i':
          ilcp = true; selected = true; break;
        case 'm':
          muthu = true; selected = true; break;
        case 'p':
          pdl = true; selected = true; break;
        case 's':
          sada = true; selected = true; break;
        default:
          std::cerr << "Invalid parameter: " << argv[arg] << std::endl;
          break;
      }
    }
    else if(base_arg == 0)    { base_arg = argv[arg]; }
    else if(pattern_arg == 0) { pattern_arg = argv[arg]; }
  }
  if(pattern_arg == 0)
  {
    printUsage();
    return 1;
  }
  if(!selected) { brute = true; }

  // Output
  std::cout << "Algorithms:";
  if(brute) { std::cout << " brute"; }
  if(ilcp)  { std::cout << " ilcp"; }
  if(ilcp)  { std::cout << " muthu"; }
  if(pdl)  { std::cout << " pdl"; }
  if(sada)  { std::cout << " sada"; }
  std::cout << std::endl;
  std::string base_name = base_arg;
  std::cout << "Base name: " << base_name << std::endl;
  std::string patterns_name = pattern_arg;
  std::cout << "Patterns: " << patterns_name << std::endl;
  std::cout << std::endl;

  // Index
  RLCSA rlcsa(base_name);
  if(!(rlcsa.isOk())) { return 2; }
  rlcsa.printInfo();
  rlcsa.reportSize(true);

  // Patterns
  std::ifstream pattern_file(patterns_name.c_str(), std::ios_base::binary);
  if(!pattern_file)
  {
    std::cerr << "Error opening pattern file!" << std::endl;
    return 3;
  }
  std::vector<std::string> rows;
  CSA::readRows(pattern_file, rows, true);
  pattern_file.close();

  // Counting
  std::vector<pair_type> ranges; ranges.reserve(rows.size());
  usint total_size = 0, total_occ = 0;
  double start = CSA::readTimer();
  for(usint i = 0; i < rows.size(); i++)
  {
    pair_type result = rlcsa.count(rows[i]);
    total_size += rows[i].length(); total_occ += CSA::length(result);
    ranges.push_back(result);
  }
  double seconds = CSA::readTimer() - start;
  double megabytes = total_size / (double)CSA::MEGABYTE;
  std::cout << "Patterns:     " << rows.size() << " (" << megabytes << " MB)" << std::endl;
  std::cout << "Occurrences:  " << total_occ << std::endl;
  std::cout << "Counting:     " << seconds << " seconds (" << (megabytes / seconds) << " MB/s, "
                                << (rows.size() / seconds) << " patterns/s)" << std::endl;
  std::cout << std::endl;
  writeRanges(ranges, patterns_name + ".ranges");

  // Brute force
  if(brute)
  {
    start = CSA::readTimer();
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++)
    {
      usint* docs = rlcsa.locate(ranges[i]);
      if(docs != 0)
      {
        usint temp = CSA::length(ranges[i]);
        rlcsa.getSequenceForPosition(docs, temp);
        CSA::sequentialSort(docs, docs + temp);
        docc += std::unique(docs, docs + temp) - docs;
        delete docs; docs = 0;
      }
    }
    seconds = CSA::readTimer() - start;
    printResults("Brute", seconds, docc, rows.size());
    printSize("Brute", rlcsa.getSize(), rlcsa.reportSize(), 0);
    std::cout << std::endl;
  }

  // ILCP
  if(ilcp)
  {
    DoclistILCP doclist(rlcsa, base_name);
    if(!(doclist.isOk()))
    {
      std::cerr << "Error: Cannot initialize DoclistILCP!" << std::endl;
      return 4;
    }
    start = CSA::readTimer();
    Doclist::found_type found(rlcsa.getNumberOfSequences(), 1);
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++)
    {
      Doclist::result_type* res = doclist.listDocuments(ranges[i], &found);
      if(res != 0)
      {
        docc += res->size();
        delete res; res = 0;
      }
    }
    seconds = CSA::readTimer() - start;
    printResults("ILCP", seconds, docc, rows.size());
    printSize("ILCP", rlcsa.getSize(), rlcsa.reportSize(), doclist.reportSize());
    std::cout << std::endl;
  }

  // Muthu
  if(muthu)
  {
    DoclistMuthu doclist(rlcsa, base_name);
    if(!(doclist.isOk()))
    {
      std::cerr << "Error: Cannot initialize DoclistMuthu!" << std::endl;
      return 5;
    }
    start = CSA::readTimer();
    Doclist::found_type found(rlcsa.getNumberOfSequences(), 1);
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++)
    {
      Doclist::result_type* res = doclist.listDocuments(ranges[i], &found);
      if(res != 0)
      {
        docc += res->size();
        delete res; res = 0;
      }
    }
    seconds = CSA::readTimer() - start;
    printResults("Muthu", seconds, docc, rows.size());
    printSize("Muthu", rlcsa.getSize(), rlcsa.reportSize(), doclist.reportSize());
    std::cout << std::endl;
  }

  if(pdl)
  {
    CSA::DocArray doclist(rlcsa, base_name);
    if(!(doclist.isOk()))
    {
      std::cerr << "Error: Cannot initialize DocArray!" << std::endl;
      return 6;
    }
    start = CSA::readTimer();
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++)
    {
      std::vector<usint>* res = doclist.listDocuments(ranges[i]);
      if(res != 0)
      {
        docc += res->size();
        delete res; res = 0;
      }
    }
    seconds = CSA::readTimer() - start;
    printResults("PDL", seconds, docc, rows.size());
    printSize("PDL", rlcsa.getSize(), rlcsa.reportSize(), doclist.reportSize());
    std::cout << std::endl;
  }

  // Sada
  if(sada)
  {
    DoclistSada doclist(rlcsa, base_name);
    if(!(doclist.isOk()))
    {
      std::cerr << "Error: Cannot initialize DoclistSada!" << std::endl;
      return 7;
    }
    start = CSA::readTimer();
    Doclist::found_type found(rlcsa.getNumberOfSequences(), 1);
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++)
    {
      Doclist::result_type* res = doclist.listDocuments(ranges[i], &found);
      if(res != 0)
      {
        docc += res->size();
        delete res; res = 0;
      }
    }
    seconds = CSA::readTimer() - start;
    printResults("Sadakane", seconds, docc, rows.size());
    printSize("Sadakane", rlcsa.getSize(), rlcsa.reportSize(), doclist.reportSize());
    std::cout << std::endl;
  }

  return 0;
}

void
printUsage()
{
  std::cout << "Usage: list_docs [algorithms] base_name patterns" << std::endl;
  std::cout << "  -b  Brute force (default)" << std::endl;
  std::cout << "  -i  ILCP" << std::endl;
  std::cout << "  -m  Muthukrishnan's algorithm" << std::endl;
  std::cout << "  -p  Precomputed document listing" << std::endl;
  std::cout << "  -s  Sadakane's algorithm" << std::endl;
  std::cout << std::endl;
}

void
printResults(const std::string& algorithm, double seconds, usint docc, usint patterns)
{
  std::cout << algorithm << ':';
  for(usint i = algorithm.length() + 1; i < 14; i++) { std::cout << ' '; }
  std::cout << docc << " documents in " << seconds << " seconds ("
            << (docc / seconds) << " docc/s, "
            << (patterns / seconds) << " patterns/s)" << std::endl;
}

void
printSize(const std::string& algorithm, usint text_size, usint index_size, usint doclist_size)
{
  std::cout << algorithm << ':';
  for(usint i = algorithm.length() + 1; i < 14; i++) { std::cout << ' '; }

  double doclist_mb = doclist_size / (double)CSA::MEGABYTE;
  double total_mb = (index_size + doclist_size) / (double)CSA::MEGABYTE;
  double doclist_bpc = CHAR_BIT * doclist_size / (double)text_size;
  double total_bpc = CHAR_BIT * (index_size + doclist_size) / (double)text_size;

  std::cout << doclist_mb << " MB / " << doclist_bpc << " bpc (total "
            << total_mb << " MB / " << total_bpc << " bpc)" << std::endl;
}

void
writeRanges(std::vector<pair_type>& ranges, const std::string& file_name)
{
  if(ranges.empty()) { return; }

  std::ofstream output(file_name.c_str(), std::ios_base::binary);
  if(!output)
  {
    std::cerr << "Error: Cannot open file " << file_name << " for writing patterns!" << std::endl;
    return;
  }

  for(std::vector<pair_type>::iterator iter = ranges.begin(); iter != ranges.end(); ++iter)
  {
    uint temp = iter->first;
    output.write((char*)&temp, sizeof(temp));
    temp = iter->second;
    output.write((char*)&temp, sizeof(temp));
  }
  output.close();
}
